"""compatibility layer to match the kkroening library.
@TODO: better docstrings"""

from typing import Tuple

from zprp_ffmpeg.BaseConnector import BaseConnector

from .FilterGraph import SinkFilter
from .FilterGraph import SourceFilter
from .FilterGraph import Stream
from .ProcessConnector import ProcessConnector


# this one is unique, creates the Stream object
def input(filename: str):
    source = SourceFilter(filename)
    return Stream().append(source)


def output(stream: Stream, filename: str):
    sink = SinkFilter(filename)
    stream.append(sink)
    return stream


# this api always uses process
def run(stream: Stream, extra_options: str) -> Tuple[str, str]:
    """Returns (stdout,stderr) tuple"""
    return ProcessConnector.run(stream, extra_options).communicate()


# this api always uses process
def run_async(stream: Stream) -> BaseConnector:
    """Returns handle to a process. Can raise an exception if script tries to terminate before ffmpeg is done."""
    return ProcessConnector.run(stream)



from typing import Optional

from zprp_ffmpeg.FilterGraph import Filter
from zprp_ffmpeg.FilterGraph import FilterOption
from zprp_ffmpeg.FilterGraph import Stream


def scale(
    graph: Stream,
    w: Optional[str] = None,
    h: Optional[str] = None,
    flags: Optional[str] = None,
    interl: Optional[bool] = None,
    size: Optional[str] = None,
    in_color_matrix: Optional[int] = None,
    out_color_matrix: Optional[int] = None,
    in_range: Optional[int] = None,
    out_range: Optional[int] = None,
    in_v_chr_pos: Optional[int] = None,
    in_h_chr_pos: Optional[int] = None,
    out_v_chr_pos: Optional[int] = None,
    out_h_chr_pos: Optional[int] = None,
    force_original_aspect_ratio: Optional[int] = None,
    force_divisible_by: Optional[int] = None,
    param0: Optional[float] = None,
    param1: Optional[float] = None,
    eval: Optional[int] = None,
):
    """Scale the input video size and/or convert the image format.
    :param str w: Output video width
    :param str h: Output video height
    :param str flags: Flags to pass to libswscale
    :param bool interl: set interlacing
    :param str size: set video size
    :param int in_color_matrix: set input YCbCr type
    :param int out_color_matrix: set output YCbCr type
    :param int in_range: set input color range
    :param int out_range: set output color range
    :param int in_v_chr_pos: input vertical chroma position in luma grid/256
    :param int in_h_chr_pos: input horizontal chroma position in luma grid/256
    :param int out_v_chr_pos: output vertical chroma position in luma grid/256
    :param int out_h_chr_pos: output horizontal chroma position in luma grid/256
    :param int force_original_aspect_ratio: decrease or increase w/h if necessary to keep the original AR
    :param int force_divisible_by: enforce that the output resolution is divisible by a defined integer when force_original_aspect_ratio is used
    :param float param0: Scaler param 0
    :param float param1: Scaler param 1
    :param int eval: specify when to evaluate expressions"""
    graph.append(
        Filter(
            command="scale",
            filter_type="AVMEDIA_TYPE_VIDEO",
            params=[
                FilterOption(name="w", value=w),
                FilterOption(name="h", value=h),
                FilterOption(name="flags", value=flags),
                FilterOption(name="interl", value=interl),
                FilterOption(name="size", value=size),
                FilterOption(name="in_color_matrix", value=in_color_matrix),
                FilterOption(name="out_color_matrix", value=out_color_matrix),
                FilterOption(name="in_range", value=in_range),
                FilterOption(name="out_range", value=out_range),
                FilterOption(name="in_v_chr_pos", value=in_v_chr_pos),
                FilterOption(name="in_h_chr_pos", value=in_h_chr_pos),
                FilterOption(name="out_v_chr_pos", value=out_v_chr_pos),
                FilterOption(name="out_h_chr_pos", value=out_h_chr_pos),
                FilterOption(name="force_original_aspect_ratio", value=force_original_aspect_ratio),
                FilterOption(name="force_divisible_by", value=force_divisible_by),
                FilterOption(name="param0", value=param0),
                FilterOption(name="param1", value=param1),
                FilterOption(name="eval", value=eval),
            ],
        )
    )
    return graph


def scale2ref(
    graph: Stream,
):
    """Scale the input video size and/or convert the image format to the given reference."""
    graph.append(Filter(command="scale2ref", filter_type="AVMEDIA_TYPE_VIDEO", params=[]))
    return graph



__version__ = "1.0.0"
# from . import longest
from ._api_compat import input
from ._api_compat import output
from ._api_compat import run
from ._api_compat import run_async
from .filters import hflip
from .generated_filters import *  # noqa: F403 this is impossible to avoid

# This is for `from xyz import *`, but also to make linter shut up
__all__ = ["input", "output", "run", "run_async", "hflip"]



"""There is no general 'Node' class, because it doesn't work well with object relations, source and sink filters are kind of orthogonal.
It slightly violates DRY, but the parameter types are different. It is what it is"""

from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import Optional
from typing import Union


class FilterType(Enum):
    VIDEO = "AVMEDIA_TYPE_VIDEO"
    AUDIO = "AVMEDIA_TYPE_AUDIO"


@dataclass
class FilterOption:
    name: str
    value: Any


class Filter:
    """Filters can have many inputs and many outputs, holds the filter name and potential params"""

    def __init__(self, command: str, params: Optional[list[FilterOption]] = None, filter_type: FilterType = FilterType.VIDEO):
        self._out: list[AnyNode] = []
        self._in: list[AnyNode] = []
        self.command = command
        self.params = params if params else []
        self.filter_type = filter_type

    def add_output(self, parent: "Filter | SinkFilter"):
        self._out.append(parent)

    def add_input(self, child: "Filter | SourceFilter"):
        self._in.append(child)

    def get_command(self):
        joined_params = ":".join(p.name + "=" + p.value for p in self.params if p.value)
        if joined_params:  # if there was no option, leave empty string
            joined_params = "=" + joined_params
        if self.filter_type == "AVMEDIA_TYPE_VIDEO":
            return "-vf " + self.command + joined_params

        elif self.filter_type == "AVMEDIA_TYPE_AUDIO":
            return "-af " + self.command + joined_params


# names as per ffmpeg documentation
class SourceFilter:
    """There can't be any input node to input filter, it provides the input itself."""

    def __init__(self, in_path: str):
        self.in_path: str = in_path
        self._out: list[AnyNode] = []

    def add_output(self, parent: "Filter | SinkFilter"):
        self._out.append(parent)

    def add_input(self, child: "Filter"):
        raise NotImplementedError("This node can't have inputs")

    def get_command(self):
        return "-i " + self.in_path


class SinkFilter:
    """Similarly to SourceFilter, it doesn't make sense to output anything further, this is the end of graph."""

    def __init__(self, out_path: str):
        self.out_path: str = out_path
        self._in: list[AnyNode] = []

    def add_input(self, parent: "Filter | SourceFilter"):
        self._in.append(parent)

    def add_output(self, parent: "Filter"):
        raise NotImplementedError("This node can't have outputs")

    def get_command(self):
        return self.out_path


# in python 3.12 there is 'type' keyword, but we are targetting 3.8
# https://stackoverflow.com/questions/76712720/typeerror-unsupported-operand-types-for-type-and-nonetype
# python >3.9 uses , instead of | idk if it works with python 3.12
AnyNode = Union[Filter, SourceFilter, SinkFilter]


class Stream:
    """One directional sequence of nodes, in future will be able to visualize them.
    Streams can be concatenated and split with certain filters."""

    def __init__(self) -> None:
        self._nodes: list[AnyNode] = []

    def append(self, node: AnyNode) -> "Stream":
        if len(self._nodes) > 0:
            # connect head with new node
            if not isinstance(self._nodes[-1], SinkFilter) and not isinstance(node, SourceFilter):
                self._nodes[-1].add_output(node)
                node.add_input(self._nodes[-1])
        self._nodes.append(node)
        return self  # fluent



import shlex
import subprocess

from .BaseConnector import BaseConnector
from .FilterGraph import Stream


class ProcessConnector(BaseConnector):
    # @TODO: this isn't ideal because attacker can sideload malicious executable...
    # Maybe warn user to set this in config for safety?
    ffmpeg_executable_path = "ffmpeg"

    def __init__(self, ffmpeg_process) -> None:
        self.ffmpeg_process: subprocess.Popen = ffmpeg_process
        super().__init__()

    @staticmethod
    def compile(graph: Stream) -> str:
        """
        Builds a command for ffmpeg from FilterGraph

        :param graph: the graph to compile
        :return: a string to pass as an argument to ffmpeg
        """

        command = [node.get_command() for node in graph._nodes]

        return " ".join(command)

    @classmethod
    def run(cls, graph: Stream, extra_options: str = "") -> "BaseConnector":
        """
        Builds a command from FilterGraph, starts ffmpeg process, and passes the command.

        :return: subprocess.Popen instance
        """

        command = ProcessConnector.compile(graph) + extra_options
        print("Command:", command)
        ffmpeg_process = subprocess.Popen([ProcessConnector.ffmpeg_executable_path, *shlex.split(command)])  # noqa: S603
        return cls(ffmpeg_process)

    def communicate(self):
        return self.ffmpeg_process.communicate()



from abc import ABC
from abc import abstractmethod

from .FilterGraph import Stream


class BaseConnector(ABC):
    """
    Abstract class for talking with ffmpeg.

    A connector always needs a `run` method to execute something user wants.
    """

    # this is basically a named constructor
    @classmethod
    @abstractmethod
    def run(cls, graph: Stream) -> "BaseConnector":
        """Executes given fliter graph
        :param graph: filter graph describing user operations.

        :return: a handle to read output from ffmpeg, for example stdout from process."""

    @abstractmethod
    def communicate(self):
        pass
